from pyqqq.utils.logger import get_logger
import datetime as dtm
import pandas as pd
import pyqqq
import pyqqq.config as c
import pytz
import requests

logger = get_logger("realtime")

def get_all_last_trades():
    '''
    모든 종목의 최근 체결 정보를 반환합니다.

    Returns:
        list:
        - dict:
            - chetime (str): 체결시간
            - sign (str): 전일대비구분
            - change (int): 전일대비가격
            - drate (float): 전일대비등락율
            - price (int): 체결가
            - opentime (str): 시가시간
            - open (int): 시가
            - hightime (str): 고가시간
            - high (int): 고가
            - lowtime (str): 저가시간
            - low (int): 저가
            - cgubun (str): 체결구분
            - cvolume (int): 체결량
            - value (int): 누적거래대금
            - mdvolume (int): 매도체결수량
            - mdchecnt (int): 매도체결건수
            - msvolume (int): 매수체결수량
            - mschecnt (int): 매수체결건수
            - cpower (float): 체결강도
            - offerho (int): 매도호가
            - bidho (int): 매수호가
            - status (str): 장정보
            - jnilvolume (int): 전일거래량
            - shcode (str): 종목코드

    '''

    r = requests.get(f'https://qupiato.com/api/domestic-stock/trades')
    r.raise_for_status()

    data = r.json()
    result = [data[k] for k in data.keys()]

    return result


def get_all_minute_data(
    time: dtm.datetime, include_empty: bool = False
) -> pd.DataFrame:
    """
    모든 종목의 분봉 데이터를 반환합니다.

    Args:
        time (dtm.datetime): 조회할 시간
        include_empty (bool): 거래가 없는 종목 데이터도 포함할지 여부 (기본값: False)

    Returns:
        pd.DataFrame: 요청한 시간의 모든 종목의 분봉 데이터를 담은 DataFrame

        - code (str): 종목 코드
        - time (dtm.datetime): 시간
        - open (int): 시가
        - high (int): 고가
        - low (int): 저가
        - close (int): 종가
        - volume (int): 거래량

    Examples:
        >>> df = get_all_minute_data(dtm.datetime(2024, 4, 2, 13, 0))
        >>> print(df)
                                time    open    high     low   close  volume
        code
        310210 2024-04-02 13:00:00   33400   33450   33400   33450      11
        000270 2024-04-02 13:00:00  104300  104500  104300  104500    7386
        359090 2024-04-02 13:00:00    1449    1456    1447    1451   21871
        003670 2024-04-02 13:00:00  288000  288000  287500  288000      91
        028300 2024-04-02 13:00:00  100400  100500  100300  100500    4701
        ...                    ...     ...     ...     ...     ...     ...
        002600 2024-04-02 13:00:00  171900  171900  171900  171900       1
        001725 2024-04-02 13:00:00   63400   63400   63400   63400      83
        005745 2024-04-02 13:00:00    9770    9770    9770    9770     400
        032685 2024-04-02 13:00:00   10090   10090   10090   10090      12
        003075 2024-04-02 13:00:00   13790   13790   13790   13790       1

        [1981 rows x 6 columns]
    """
    tz = pytz.timezone("Asia/Seoul")

    url = f"{c.PYQQQ_API_URL}/domestic-stock/ohlcv/minutes/all/{time.date()}/{time.strftime('%H%M')}"
    if include_empty:
        url += "?includeEmpty=true"

    r = requests.get(url, headers={"Authorization": f"Bearer {pyqqq.get_api_key()}"})
    if r.status_code != 200 and r.status_code != 201:
        logger.error(f"Failed to get minute data: {r.text}")
        return

    rows = r.json()
    for data in rows:
        time = (
            dtm.datetime.fromisoformat(data["time"]).astimezone(tz).replace(tzinfo=None)
        )
        data["time"] = time

    df = pd.DataFrame(rows)
    if not df.empty:
        df.set_index("code", inplace=True)

    return df



import time
from functools import wraps

def retry(exceptions, total_tries=5, delay=0.5, backoff=2, silently: bool = False):
    """
    Decorator for retrying function if exception occurs

    Args:
        exceptions (Exception or tuple): exception(s) to check. can be a tuple of exceptions
        total_tries (int): total tries to attempt
        delay (float): initial delay between retry in seconds
        backoff (float): backoff multiplier e.g. value of 2 will double the delay each retry
        silently (bool): if True, will not print exception message
    Returns:
        function: Decorated function that will retry upon exceptions
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            current_try = 0
            while current_try < total_tries:
                try:
                    return func(*args, **kwargs)
                except exceptions as e:
                    current_try += 1
                    sleep_time = delay * (backoff ** (current_try - 1))
                    if current_try != total_tries:
                        if not silently:
                            print(f"{str(e)}\nRetrying in {sleep_time} seconds...")
                        time.sleep(sleep_time)
                    else:
                        print("Max retry attempts reached, aborting.")
                        raise
        return wrapper
    return decorator


from dataclasses import dataclass
from decimal import Decimal
from pyqqq.brokerage.kis.domestic_stock import KISDomesticStock
from pyqqq.brokerage.kis.oauth import KISAuth
from pyqqq.data.realtime import get_all_last_trades
from pyqqq.datatypes import *
from pyqqq.utils.market_schedule import get_market_schedule
from typing import List, Optional
import datetime as dtm
import pandas as pd


@dataclass
class KISStockPosition:
    """
    주식 보유 종목 정보를 담는 데이터 클래스입니다.
    """

    asset_code: str
    """ 종목 코드 """
    asset_name: str
    """ 종목 이름 """
    quantity: int
    """ 보유 수량 """
    average_purchase_price: Decimal
    """ 평균 매입 가격 """
    current_price: int
    """ 현재 가격 """
    current_value: int
    """ 현재 가치 """
    current_pnl: Decimal
    """ 현재 손익률 """


@dataclass
class KISStockOrder:
    """
    주식 주문 정보를 담는 데이터 클래스입니다.
    """

    order_no: str
    """ 주문 번호 """
    asset_code: str
    """ 종목 코드 """
    side: OrderSide
    """ 주문 방향 """
    quantity: int
    """ 주문 수량 """
    price: int
    """ 주문 가격 """
    filled_quantity: int
    """ 체결 수량 """
    filled_price: int
    """ 체결 가격 """
    pending_quantity: int
    """ 미체결 수량 """
    order_time: dtm.datetime
    """ 주문 시각 """
    current_price: int
    """ 현재 가격 """
    is_pending: Optional[bool] = None
    """ 미체결 여부 """
    org_order_no: Optional[str] = None
    """ 원주문번호 """
    order_type: Optional[OrderType] = None
    """ 주문 유형 """
    req_type: Optional[OrderRequestType] = None
    """ 주문 요청 유형 """


class KISSimpleDomesticStock:

    """
    한국투자증권 국내 주식 API 사용하여 주식 거래를 하기 위한 클래스입니다.

    기존 KISDomesticStock 클래스를 감싸고, 간단한 주문/조회 기능을 제공합니다.

    Attributes:
        auth (KISAuth): 인증 정보
        account_no (str): 계좌 번호
        account_product_code (str): 계좌 상품 코드
    """
    def __init__(self,
                 auth: KISAuth,
                 account_no: str,
                 account_product_code: str):

        self.account_no = account_no
        self.account_product_code = account_product_code
        self.stock_api = KISDomesticStock(auth)

    def get_account(self) -> dict:
        """
        계좌 정보를 조회합니다.

        Returns:
            dict: 계좌 정보

            - total_balance (int): 총 평가 금액
            - purchase_amount (int): 매입 금액
            - evaluated_amount (int): 평가 금액
            - pnl_amount (int): 손익 금액
            - pnl_rate (Decimal): 손익률
        """
        r = self.stock_api.inquire_balance(self.account_no, self.account_product_code, '01')
        data = r['output2'][0]

        purchase_amount = int(data["pchs_amt_smtl_amt"])
        pnl_amount = int(data["evlu_pfls_smtl_amt"])
        pnl_rate = (
            Decimal(pnl_amount / purchase_amount)
            if purchase_amount != 0
            else Decimal(0)
        )

        result = {
            "total_balance": int(data["tot_evlu_amt"]),
            "purchase_amount": purchase_amount,
            "evaluated_amount": int(data["evlu_amt_smtl_amt"]),
            "pnl_amount": pnl_amount,
            "pnl_rate": pnl_amount,
        }
        return result

    def get_possible_quantity(self,
                              asset_code: str,
                              order_type: OrderType = OrderType.MARKET,
                              price: int = 0) -> dict:
        """
        주문 가능 수량을 조회합니다.

        Args:
            asset_code (str): 종목 코드
            order_type (OrderType): 주문 유형
            price (int): 주문 가격 (지정가 주문일 경우에만 필요)

        Returns:
            dict: 주문 가능 수량 정보

            - investable_cash (int): 주문 가능 현금
            - reusable_amount (int): 재사용 가능 금액
            - price (int): 계산 기준 단가
            - quantity (int): 주문 가능 수량
            - amount (int): 주문 시 소요 금액
        """

        def __get_order_type_code():
            if order_type == OrderType.MARKET:
                return "01"
            elif order_type == OrderType.LIMIT:
                return "00"
            elif order_type == OrderType.LIMIT_IOC:
                return "11"
            elif order_type == OrderType.LIMIT_FOK:
                return "12"
            elif order_type == OrderType.MARKET_IOC:
                return "13"
            elif order_type == OrderType.MARKET_FOK:
                return "14"
            else:
                raise ValueError("지원하지 않는 주문 유형입니다.")

        resp = self.stock_api.inquire_psbl_order(
            self.account_no,
            self.account_product_code,
            asset_code,
            __get_order_type_code(),
            price,
        )

        data = resp['output']
        result = {
            "investable_cash": data["ord_psbl_cash"],
            "reusable_amount": data["ruse_psbl_amt"],
            "price": data["psbl_qty_calc_unpr"],
            "quantity": data["max_buy_qty"],
            "amount": data["max_buy_amt"],
        }
        return result

    def get_positions(self) -> List[KISStockPosition]:
        """
        보유 종목을 조회합니다.

        Returns:
            List[KISStockPosition]: 보유 종목 정보 리스트
        """

        fetching = True
        ctx_area_fk100 = ""
        ctx_area_nk100 = ""
        tr_cont = ""

        result = []

        while fetching:
            r = self.stock_api.inquire_balance(
                self.account_no,
                self.account_product_code,
                "02",
                tr_cont=tr_cont,
                ctx_area_fk100=ctx_area_fk100,
                ctx_area_nk100=ctx_area_nk100,
            )

            for el in r["output1"]:
                position = KISStockPosition(
                    asset_code=el["pdno"],
                    asset_name=el["prdt_name"],
                    quantity=int(el["hldg_qty"]),
                    average_purchase_price=Decimal(el["pchs_avg_pric"]),
                    current_price=int(el["prpr"]),
                    current_value=int(el["evlu_amt"]),
                    current_pnl=Decimal(el["evlu_pfls_rt"]),
                )

                if position.quantity > 0:
                    result.append(position)

            if r["tr_cont"] in ["F", "M"]:
                tr_cont = "N"
                ctx_area_fk100 = r["ctx_area_fk100"]
                ctx_area_nk100 = r["ctx_area_nk100"]
            else:
                fetching = False

        return result

    def get_historical_daily_data(self,
                                  asset_code: str,
                                  first_date: dtm.date,
                                  last_date: dtm.date,
                                  adjusted_price: bool = False
    ) -> pd.DataFrame:
        """
        일봉 데이터 검색

        Args:
            asset_code(str): 종목코드
            first_date(datetime.date): 조회 시작일자
            last_date(datetime.date): 조회 종료일자
            adjusted_price(bool): 수정 주가 여부

        Returns:
            pd.DataFrame: 일봉 데이터
        """
        assert first_date <= last_date, 'last_date는 first_date와 같거나, 이후 날짜여야 합니다'
        assert last_date <= dtm.date.today(), 'last_date는 오늘과 같거나 이전이어야 합니다.'

        max_days_per_request = 100
        total_days = (last_date - first_date).days

        result = []

        for i in range(0, total_days + 1, max_days_per_request):
            search_start = first_date + dtm.timedelta(days=i)
            search_end = min(first_date + dtm.timedelta(days=i + max_days_per_request), last_date)

            r = self.stock_api.inquire_daily_itemchartprice(
                asset_code,
                search_start,
                search_end,
                fid_period_div_code='D',
                fid_org_adj_prc='0'
                if adjusted_price else '1')

            chunk = []
            for item in r['output2']:
                if not item:
                    continue

                chunk.append({
                    'date': item['stck_bsop_date'],
                    'open': item['stck_oprc'],
                    'high': item['stck_hgpr'],
                    'low': item['stck_lwpr'],
                    'close': item['stck_clpr'],
                    'volume': item['acml_vol'],
                })
            chunk.reverse()
            result.extend(chunk)

        df = pd.DataFrame(result)
        df['date'] = pd.to_datetime(df['date'])
        df.set_index('date', inplace=True)

        return df

    def get_today_minute_data(self, asset_code: str) -> pd.DataFrame:
        """
        분봉 데이터 검색

        Args:
            asset_code(str): 종목코드

        Returns:
            pd.DataFrame: 분봉 데이터
        """

        request_time = dtm.datetime.now().replace(second=0, microsecond=0)
        result = []
        schedule = get_market_schedule(dtm.date.today())

        while True:
            r = self.stock_api.inquire_time_itemchartprice(asset_code, request_time.time(), fid_pw_data_incu_yn='N')
            output = r['output2']
            if len(output) == 0:
                break

            for item in output:
                if not item:
                    continue

                result.append({
                    'time': dtm.datetime.combine(item['stck_bsop_date'], item['stck_cntg_hour']),
                    'open': item['stck_oprc'],
                    'high': item['stck_hgpr'],
                    'low': item['stck_lwpr'],
                    'close': item['stck_prpr'],
                    'volume': item['cntg_vol'],
                })

            last_item_time = result[-1]['time']
            request_time = last_item_time - dtm.timedelta(minutes=1)
            if request_time.time() < schedule.open_time:
                break

        df = pd.DataFrame(result)
        df['time'] = pd.to_datetime(df['time'], format='%H:%M:%S')
        df.set_index('time', inplace=True)

        return df

    def get_price(self, asset_code: str) -> dict:
        """
        주식 현재 가격 조회

        Args:
            asset_code(str): 종목코드

        Returns:
            dict: 현재 가격 정보

            - code (str): 종목 코드
            - current_price (int): 현재 가격
            - volume (int): 거래량
            - open_price (int): 시가
            - high_price (int): 고가
            - low_price (int): 저가
            - max_price (int): 상한가
            - min_price (int): 하한가
        """
        r = self.stock_api.get_price(asset_code)

        data = r['output']
        result = {
            'code': asset_code,
            'current_price': data['stck_prpr'],
            'volume': data['acml_vol'],
            'open_price': data['stck_oprc'],
            'high_price': data['stck_hgpr'],
            'low_price': data['stck_lwpr'],
            'max_price': data['stck_mxpr'],
            'min_price': data['stck_llam'],
        }

        return result

    def get_price_for_multiple_stock(self, asset_codes: List[str]) -> pd.DataFrame:
        """
        여러 종목의 현재 가격 조회

        Args:
            asset_codes(List[str]): 종목 코드 리스트

        Returns:
            pd.DataFrame: 현재 가격 정보

            - code (str): 종목 코드
            - current_price (int): 현재 가격
            - volume (int): 거래량
            - open_price (int): 시가
            - high_price (int): 고가
            - low_price (int): 저가
        """
        r = get_all_last_trades()
        result = []

        for item in r:
            if item['shcode'] in asset_codes:
                result.append({
                    'code': item['shcode'],
                    'current_price': item['price'],
                    'volume': item['volume'],
                    'open_price': item['open'],
                    'high_price': item['high'],
                    'low_price': item['low'],
                })

        df = pd.DataFrame(result)
        df.set_index('code', inplace=True)
        return df

    def create_order(self,
                     asset_code: str,
                     side: OrderSide,
                     quantity: int,
                     order_type: OrderType = OrderType.MARKET,
                     price: int = 0) -> str:

        """
        주문을 생성합니다.

        Args:
            asset_code (str): 종목 코드
            side (OrderSide): 주문 방향
            quantity (int): 주문 수량
            order_type (OrderType): 주문 유형
            price (int): 주문 가격 (지정가 주문일 경우에만 필요)

        Returns:
            str: 주문 번호
        """

        def __get_order_type_code():
            if order_type == OrderType.LIMIT:
                return "00"
            elif order_type == OrderType.MARKET:
                return "01"
            elif order_type == OrderType.LIMIT_CONDITIONAL:
                return "02"
            elif order_type == OrderType.BEST_PRICE:
                return "03"
            elif order_type == OrderType.PRIMARY_PRICE:
                return "04"
            elif order_type == OrderType.LIMIT_IOC:
                return "11"
            elif order_type == OrderType.LIMIT_FOK:
                return "12"
            elif order_type == OrderType.MARKET_IOC:
                return "13"
            elif order_type == OrderType.MARKET_FOK:
                return "14"
            elif order_type == OrderType.BEST_PRICE_IOC:
                return "15"
            elif order_type == OrderType.BEST_PRICE_FOK:
                return "16"
            else:
                raise ValueError("지원하지 않는 주문 유형입니다.")

        def __get_order_side_code():
            if side == OrderSide.BUY:
                return "buy"
            elif side == OrderSide.SELL:
                return "sell"
            else:
                raise ValueError("지원하지 않는 주문 방향입니다.")

        r = self.stock_api.order_cash(
            self.account_no,
            self.account_product_code,
            __get_order_side_code(),
            asset_code,
            __get_order_type_code(),
            quantity,
            price,
        )

        return r['output']['ODNO']

    def update_order(self,
                     org_order_no: str,
                     order_type: OrderType,
                     price: int,
                     quantity: int = 0) -> str:

        """
        주문을 수정합니다.

        Args:
            org_order_no (str): 원주문번호
            order_type (OrderType): 주문 유형
            price (int): 정정 가격
            quantity (int): 주문 수량

        Returns:
            str: 주문 번호
        """

        def __get_order_type_code():
            if order_type == OrderType.LIMIT:
                return "00"
            elif order_type == OrderType.MARKET:
                return "01"
            elif order_type == OrderType.LIMIT_CONDITIONAL:
                return "02"
            elif order_type == OrderType.BEST_PRICE:
                return "03"
            elif order_type == OrderType.PRIMARY_PRICE:
                return "04"
            elif order_type == OrderType.LIMIT_IOC:
                return "11"
            elif order_type == OrderType.LIMIT_FOK:
                return "12"
            elif order_type == OrderType.MARKET_IOC:
                return "13"
            elif order_type == OrderType.MARKET_FOK:
                return "14"
            elif order_type == OrderType.BEST_PRICE_IOC:
                return "15"
            elif order_type == OrderType.BEST_PRICE_FOK:
                return "16"
            else:
                raise ValueError("지원하지 않는 주문 유형입니다.")

        qty_all_ord_yn = 'Y' if quantity == 0 else 'N'

        r = self.stock_api.order_rvsecncl(
            self.account_no,
            self.account_product_code,
            org_order_no,
            __get_order_type_code(),
            '01', # 정정
            quantity,
            price,
            qty_all_ord_yn
        )

        return r['output']['ODNO']

    def cancel_order(self, org_order_no: str, quantity: int = 0) -> str:
        """
        주문을 취소합니다.

        Args:
            org_order_no (str): 원주문번호
            quantity (int): 취소 수량 (일부 취소의 경우 지정)
        """
        r = self.stock_api.order_rvsecncl(
            self.account_no,
            self.account_product_code,
            org_order_no,
            '00',
            '02', # 취소
            quantity,
            0,
            'Y' if quantity == 0 else 'N'
        )

        return r['output']['ODNO']

    def get_pending_orders(self) -> List[KISStockOrder]:
        """
        미체결 주문을 조회합니다.

        Returns:
            List[KISStockOrder]: 미체결 주문 리스트
        """
        fetching = True
        ctx_area_fk100 = ""
        ctx_area_nk100 = ""

        result: List[KISStockOrder] = []
        asset_codes = []

        while fetching:
            r = self.stock_api.inquire_psbl_rvsecncl(
                self.account_no,
                self.account_product_code,
                ctx_area_fk100=ctx_area_fk100,
                ctx_area_nk100=ctx_area_nk100,
            )

            for item in r["output"]:
                filled_amount = int(item.get("tot_ccld_amt", 0))
                filled_quantity = int(item.get("tot_ccld_qty", 0))
                filled_price = (
                    filled_amount // filled_quantity if filled_quantity > 0 else 0
                )
                asset_code = item["pdno"]
                side = (
                    OrderSide.SELL if item["sll_buy_dvsn_cd"] == "01" else OrderSide.BUY
                )

                order = KISStockOrder(
                    order_no=item["odno"],
                    asset_code=asset_code,
                    side=side,
                    price=int(item["ord_unpr"]),
                    filled_price=filled_price,
                    current_price=0,
                    quantity=int(item["ord_qty"]),
                    filled_quantity=filled_quantity,
                    pending_quantity=item["psbl_qty"],
                    order_time=item["ord_tmd"],
                )

                asset_codes.append(asset_code)
                result.append(order)

            if r["tr_cont"] in ["F", "M"]:
                ctx_area_fk100 = r["ctx_area_fk100"]
                ctx_area_nk100 = r["ctx_area_nk100"]
            else:
                fetching = False

        if len(asset_codes) > 0:
            price_df = self.get_price_for_multiple_stock(asset_codes)

        for order in result:
            order.current_price = int(price_df.loc[order.asset_code, "current_price"])

        return result

    def get_today_order_history(
        self, target_date: dtm.date = None
    ) -> List[KISStockOrder]:
        """
        오늘 주문 내역을 조회합니다.

        Returns:
            List[KISStockOrder]: 오늘 주문 내역 리스트
        """
        result = []
        fetching = True
        ctx_area_fk100 = ""
        ctx_area_nk100 = ""
        tr_cont = ""

        pending_order_list = [item.order_no for item in self.get_pending_orders()]

        def __get_order_side_code(sll_buy_dvsn_cd: str) -> OrderSide:
            if sll_buy_dvsn_cd == "01":
                return OrderSide.SELL
            elif sll_buy_dvsn_cd == "02":
                return OrderSide.BUY
            else:
                raise ValueError("지원하지 않는 주문 방향입니다.")

        def __get_req_type(sll_buy_dvsn_cd_name: str) -> OrderRequestType:
            if "취소" in sll_buy_dvsn_cd_name:
                return OrderRequestType.CANCEL
            elif "정정" in sll_buy_dvsn_cd_name:
                return OrderRequestType.MODIFY
            else:
                return OrderRequestType.NEW

        def __get_order_type(ord_dvsn_cd: str) -> OrderType:
            if ord_dvsn_cd == "00":
                return OrderType.LIMIT
            elif ord_dvsn_cd == "01":
                return OrderType.MARKET
            elif ord_dvsn_cd == "02":
                return OrderType.LIMIT_CONDITIONAL
            elif ord_dvsn_cd == "03":
                return OrderType.BEST_PRICE
            elif ord_dvsn_cd == "04":
                return OrderType.PRIMARY_PRICE
            elif ord_dvsn_cd == "11":
                return OrderType.LIMIT_IOC
            elif ord_dvsn_cd == "12":
                return OrderType.LIMIT_FOK
            elif ord_dvsn_cd == "13":
                return OrderType.MARKET_IOC
            elif ord_dvsn_cd == "14":
                return OrderType.MARKET_FOK
            elif ord_dvsn_cd == "15":
                return OrderType.BEST_PRICE_IOC
            elif ord_dvsn_cd == "16":
                return OrderType.BEST_PRICE_FOK
            else:
                raise ValueError(f"지원하지 않는 주문 유형입니다. {ord_dvsn_cd}")

        if target_date is None:
            target_date = dtm.date.today()

        while fetching:
            r = self.stock_api.inquire_daily_ccld(
                self.account_no,
                self.account_product_code,
                inqr_strt_dt=target_date,
                inqr_end_dt=target_date,
                ctx_area_fk100=ctx_area_fk100,
                ctx_area_nk100=ctx_area_nk100,
                tr_cont=tr_cont,
            )

            for item in r["output1"]:
                item["pending"] = item["odno"] in pending_order_list

                quantity = int(item["ord_qty"])
                filled_quantity = int(item["tot_ccld_qty"])
                filled_amount = int(item["tot_ccld_amt"])
                pending_quantity = quantity - filled_quantity
                filled_price = (
                    filled_amount // filled_quantity if filled_quantity > 0 else 0
                )

                order = KISStockOrder(
                    order_no=item["odno"],
                    asset_code=item["pdno"],
                    side=__get_order_side_code(item["sll_buy_dvsn_cd"]),
                    quantity=quantity,
                    price=int(item["ord_unpr"]),
                    filled_quantity=filled_quantity,
                    filled_price=filled_price,
                    pending_quantity=pending_quantity,
                    order_time=item["ord_tmd"],
                    current_price=0,
                    is_pending=item["odno"] in pending_order_list,
                    org_order_no=item["orgn_odno"],
                    order_type=__get_order_type(item["ord_dvsn_cd"]),
                    req_type=__get_req_type(item["sll_buy_dvsn_cd_name"]),
                )
                result.append(order)

            if r["tr_cont"] in ["F", "M"]:
                ctx_area_fk100 = r["ctx_area_fk100"]
                ctx_area_nk100 = r["ctx_area_nk100"]
                tr_cont = "N"
            else:
                fetching = False

        return result



import datetime as dtm
import requests

class EBestAuth:
    '''
    이베스트투자증권 인증 정보를 담고 있는 객체

    Args:
        appkey (str): 앱 키
        appsecret (str): 앱 시크릿
        paper_trading (bool): 모의 투자 여부
    '''
    def __init__(self, appkey: str, appsecret: str, paper_trading: bool = False):
        self.appkey = appkey
        self.appsecret = appsecret
        self.paper_trading = paper_trading

        self.access_token = None
        self.expired_at: dtm.datetime = None

    @property
    def host_url(self) -> str:
        '''
        모의/실전 투자 환경에 따른 API 호스트 URL을 반환합니다.

        Returns:
            str: API 호스트 URL
        '''
        return 'https://openapi.ebestsec.co.kr:8080'

    @property
    def websocket_url(self) -> str:
        '''
        실시간 데이터 스트리밍을 위한 웹소켓 URL을 반환합니다.

        Returns:
            str: 웹소켓 URL
        '''
        if self.paper_trading:
            return 'wss://openapi.ebestsec.co.kr:29443'
        else:
            return 'wss://openapi.ebestsec.co.kr:9443'

    def get_token(self) -> str:
        '''
        액세스 토큰을 반환합니다.

        Returns:
            str: 액세스 토큰
        '''
        if self.access_token is None or self.expired_at < dtm.datetime.now():
            token, expired_at = self.issue_access_token()

            self.access_token = token
            self.expired_at = expired_at

        return self.access_token

    def issue_access_token(self):
        '''
        액세스 토큰을 발급합니다.

        Returns:
            tuple: 액세스 토큰과 만료 시간을 담은 튜플
        '''
        if self.appkey is None or self.appsecret is None:
            raise ValueError('App key and app secret must be set')

        headers = {
            'content-type': 'application/x-www-form-urlencoded'
        }
        body = {
            'grant_type': 'client_credentials',
            'appkey': self.appkey,
            'appsecretkey': self.appsecret,
            'scope': 'oob',
        }

        r = requests.post(f"{self.host_url}/oauth2/token", headers=headers, data=body)
        r.raise_for_status()

        data = r.json()

        access_token = data['access_token']
        expires_in = data['expires_in']

        next_day_am7 = dtm.datetime.now().replace(hour=7, minute=0, second=0, microsecond=0) + dtm.timedelta(days=1)
        expired_at = min(dtm.datetime.now() + dtm.timedelta(seconds=expires_in), next_day_am7)

        return access_token, expired_at



import logging
from logging.handlers import TimedRotatingFileHandler
import pyqqq.config as c
import sys

_format = "%(levelname).1s %(name)s: %(message)s"

if not c.is_google_cloud_logging_enabled():
    _format = "%(asctime)s " + _format

_stdout_h = logging.StreamHandler(sys.stdout)
_stdout_h.setLevel(logging.DEBUG)
_stdout_h.addFilter(lambda r: r.levelno <= logging.WARNING)

_stderr_h = logging.StreamHandler(sys.stderr)
_stderr_h.setLevel(logging.ERROR)

logging.basicConfig(format=_format, handlers=[_stdout_h, _stderr_h])


def get_logger(
        name,
        level=logging.DEBUG,
        filename: str = None,
        when: str = 'h',
        interval: int = 1,
        backup_count: int = 24,
) -> logging.Logger:

    logger = logging.getLogger(name)
    logger.setLevel(level)

    if not logger.handlers and filename:
        fh = TimedRotatingFileHandler(
            filename, when=when, backupCount=backup_count, interval=interval
        )
        fh.setLevel(level)
        fh.setFormatter(_format)
        logger.addHandler(fh)

    return logger



from pyqqq import get_api_key
from pyqqq.utils.retry import retry
import datetime as dtm
import pandas as pd
import pyqqq.config as c
import requests


def get_alert_stocks(alert_type: str, date: dtm.date = None) -> pd.DataFrame | None:
    """
    시장 경보 종목을 조회합니다.

    Args:
        alert_type (str): 경보종류. caution:투자주의종목 warning:투자경고종목 risk:투자위험종목
        date (dtm.date, optional): 조회할 날짜. 기본값은 None (가장 최근 데이터)

    Returns:
        pd.DataFrame|None: 경보 종목 리스트. 해당일에 저장된 데이터가 없으면 None, 저장되었지만 지정 종목이 없으면 빈 DataFrame이 반환됩니다.

        - code (str, index): 종목코드
        - name (str): 종목명
        - current_price (int): 현재가
        - change (int): 전일대비가격
        - change_rate (float): 전일대비등락율
        - volume (int): 거래량
        - bid_price (int): 매수호가
        - ask_price (int): 매도호가
        - per (float): PER

    Examples:
        >>> df = get_alert_stocks('caution')
        >>> print(df.head())
                    name  current_price  change  change_rate   volume  bid_price  ask_price    per
        code
        402340   SK스퀘어          77600    2400        -3.00   303590      77000      76900  -8.49
        053950    경남제약           1570       1         0.06  4857452       1569       1568  -4.91
        012320  경동인베스트         113100   11800        11.65   994188     118800     118700  22.05
        002720    국제약품           6820     310        -4.35  5559738       6900       6890 -17.14
        219420  링크제니시스           9100     160         1.79  1720993       9120       9100  83.49
    """
    url = f"{c.PYQQQ_API_URL}/domestic-stock/alert-stocks/{alert_type}"
    params = {}
    if date:
        params["date"] = date

    r = _send_request("GET", url, params=params)
    if r.status_code == 404:
        return None
    else:
        _raise_for_status(r)

        df = pd.DataFrame(r.json())
        df.set_index("code", inplace=True)
        return df


def get_management_stocks(date: dtm.date = None) -> pd.DataFrame | None:
    """
    관리종목을 조회합니다.

    Args:
        date (dtm.date, optional): 조회할 날짜(지정일이 아닌 데이터 수집일). 기본값은 None (가장 최근 데이터)

    Returns:
        pd.DataFrame|None: 관리종목 리스트. 해당일에 저장된 데이터가 없으면 None, 저장되었지만 지정 종목이 없으면 빈 DataFrame이 반환됩니다.

        - code (str, index): 종목코드
        - name (str): 종목명
        - current_price (int): 현재가
        - change (int): 전일대비가격
        - change_rate (float): 전일대비등락율
        - volume (int): 거래량
        - designation_date (str): 지정일
        - designation_reason (str): 지정사유

    Examples:
        >>> df = get_management_stocks()
        >>> print(df.head())
                   name  current_price  change  change_rate  volume designation_date designation_reason
        code
        001140       국보           2110       0         0.00       0       2024.03.22          감사의견 의견거절
        006380      카프로            732       0         0.00       0       2024.03.22          감사의견 의견거절
        093230     이아이디           1392       0         0.00       0       2024.03.22          감사의견 의견거절
        363280   티와이홀딩스           3205     150        -4.47  393547       2024.03.22  감사범위제한으로인한 감사의견한정
        36328K  티와이홀딩스우           4940     560       -10.18   26011       2024.03.22  감사범위제한으로인한 감사의견한정
    """
    url = f"{c.PYQQQ_API_URL}/domestic-stock/management-stocks"
    params = {}
    if date:
        params["date"] = date

    r = _send_request("GET", url, params=params)
    if r.status_code == 404:
        return None
    else:
        _raise_for_status(r)

        df = pd.DataFrame(r.json())
        df.set_index("code", inplace=True)
        return df


@retry(requests.HTTPError)
def _send_request(method: str, url: str, **kwargs):
    api_key = get_api_key()
    if not api_key:
        raise ValueError("API key is not set")

    return requests.request(
        method=method,
        url=url,
        headers={"Authorization": f"Bearer {api_key}"},
        **kwargs,
    )


def _raise_for_status(r: requests.Response):
    if r.status_code != 200:
        print(r.text)

    r.raise_for_status()



import os
import pyqqq.config as c

_api_key = None

def set_api_key(api_key: str):
    global _api_key

    assert api_key is not None, "API key must not be None"
    assert len(api_key) >= 32, "API key must be at least 32 characters long"

    _api_key = api_key


def get_api_key() -> str | None:
    if _api_key:
        return _api_key

    elif c.get_pyqqq_api_key() is not None:
        return c.get_pyqqq_api_key()

    elif os.path.exists(c.get_credential_file_path()):
        with open(c.get_credential_file_path(), "r") as f:
            return f.read().strip()

    return None



import time


_MAX_CALLS = 5
_PERIOD = 1
_SCOPE = 'default'


class CallLimiter:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(CallLimiter, cls).__new__(cls)
            cls._instance.windows = {'default': []}

        return cls._instance


    def wait_limit_rate(self, max_calls: int = _MAX_CALLS, period: float = _PERIOD, scope: str = _SCOPE):
        if scope not in self.windows:
            self.windows[scope] = []

        now = time.time()
        window = list(filter(lambda x: now - x < period, self.windows[scope]))
        window_size = len(window)

        if window_size >= max_calls:
            time.sleep(period)
        elif window_size > 0:
            time.sleep(period / max_calls)

        window.append(time.time())
        self.windows[scope] = window


def limit_calls(max_calls: int = _MAX_CALLS, period: float = _PERIOD, scope: str = _SCOPE):
    """
    여러 함수에 대한 호출을 제한하는 데코레이터입니다.

    Args:
        max_calls (int): 최대 호출 횟수
        period (float): 기간(초) 동안의 호출 횟수를 제한합니다.
        scope (str): 호출 제한을 적용할 스코프 이름입니다.
    """
    def decorator(func):
        def wrapper(*args, **kwargs):
            CallLimiter().wait_limit_rate(max_calls=max_calls, period=period, scope=scope)
            result = func(*args, **kwargs)
            return result

        return wrapper

    return decorator